﻿using System;
using System.Collections.Generic;	//List
using System.IO;	//MemoryStream
using System.Threading;		//Sleep
using ModifiedSerialPort;

namespace I2cEepRomTool
{
	//通信コマンド（文字）
	enum COMCmd
	{					//引数
		DEVINFO	= '(',	//EEPROM基本情報：チップセレクトビット1byte, メモリ終端アドレス2byte, ページサイズ1byte
		STARTDATE	= ')',	//測定開始日時6byte：西暦4桁 2byte[Hi][Lo], 何月何日2byte, 何時何分2byte
		WRITE	= '#',	//アドレス2byte[Hi][Lo], データの個数2byte[Hi][Lo], データ（nバイト）	※データの個数はDATA_BUF_SIZEを超えないこと
		READ	= '$',	//アドレス2byte[Hi][Lo], データの個数2byte[Hi][Lo]
		ERASE	= '[',	//メモリ領域を埋める値1byte（EEPROM消去）
		WHO		= '@',	//なし（デバイス名の確認）

		RET_UNKNOWN_CMD = '?',	//解釈できないコマンドを受信した場合に送り返す文字（デバイスが）
		RET_ERROR	= RET_UNKNOWN_CMD	//読み書き処理が失敗したなどエラー発生の合図として返す（デバイスが）
	};

	class EepRomCtrl
	{
		//インターフェイスのデバイスに埋め込まれている名前
		private const string DEVICE_NAME = "I2CEEPROM";

		//インターフェイスのデバイスのバッファに格納できるデータ長(byte)
		private const int DATA_BUF_SIZE = 512;
		
		//1回の通信の終了コード
		private const char END_CODE = '\n';
		private static readonly char[] TXD_END_CODE = { END_CODE };

		private MSerialPort comCtrl = null;
		public static bool IsComError { get; private set; }	//通信エラーを通知するフラグ

		//コンストラクタ
		public EepRomCtrl()
		{
			IsComError = false;
		}
		
		//COMポートオブジェクトをセットする
		public void SetSerialPort(MSerialPort msp)
		{
			comCtrl = msp;
		}

		//指定されたCOMポートにデバイスが接続されているか確認する
		public bool DeviceContact(MSerialPort testComPort)
		{
			if (testComPort == null || !testComPort.IsOpen) return false;
			MSerialPort orgComCtrl = comCtrl;
			comCtrl = testComPort;

			Command(COMCmd.WHO);
			string s = ReadString();
			bool isOK = (!IsComError && s == DEVICE_NAME);

			comCtrl = orgComCtrl;
			return isOK;
		}
		
		//引数の下位4bitの数値0～15にゲタ'A'を履かせて文字に変換する
		private byte itoc(int n) { return (byte)('A' + (n & 0x0F)); }

		//ゲタ'A'を履かせた文字から元の数値0～15に変換する
		private byte ctoi(byte c) { return (byte)(c - 'A'); }

		//コマンドとパラメータを送信する
		//・第2引数は可変長配列。または省略可。
		//・データは、1byteの値を上位／下位4bitに分け、それを加工した文字として送る。
		private void Command(COMCmd cmd, params byte[] values)
		{
			IsComError = false;
			byte[] byteArray = null;
			
			using (MemoryStream memStream = new MemoryStream())
			{
				memStream.WriteByte((byte)cmd);
				
				for (int i = 0; i < values.Length; i++)
				{
					memStream.WriteByte(itoc(values[i] >> 4));
					memStream.WriteByte(itoc(values[i]));
				}
				byteArray = memStream.ToArray();
			}

			try
			{
				RecieveBufClear();	//受信に備え、PC側の受信バッファをクリアしておく
				comCtrl.Write(byteArray, 0, byteArray.Length);
				comCtrl.Write(TXD_END_CODE, 0, TXD_END_CODE.Length);
			}
			catch { IsComError = true; }
		}

		//PC側の受信バッファをクリアする
		private void RecieveBufClear()
		{
			comCtrl.ReadExisting();
			Thread.Sleep(20);	//バッファクリアを確実にするため、FT232Rのレイテンシー時間以上待つ
		}

		//送信結果を文字列として受信する
		private string ReadString()
		{
			IsComError = false;
			string s = null;
			try { s = comCtrl.ReadLine().Trim(); }	//Trim()は保険
			catch { IsComError = true; }	//タイムアウト発生はキャッチされる
			return s;
		}
		
		//書き込み系コマンドに対するデバイスからの返信の妥当性をチェックする
		//・書き込み系コマンドを送信すると、コマンド自身を表す文字が返ってくる。
		private bool WriteCommandIsDone(COMCmd cmd)
		{
			string sRetData = ReadString();
			if (IsComError) return false;
			string sWrCmd = ((char)cmd).ToString();
			return (sWrCmd == sRetData);
		}

		//コマンドを送信する
		private bool SendCommand(COMCmd cmd, int adrs, int length, params byte[] values)
		{
			byte[] byteArray = { HiByte(adrs), LoByte(adrs), HiByte(length), LoByte(length) };
			if (0 < values.Length)
			{
				using (MemoryStream memStream = new MemoryStream())
				{
					memStream.Write(byteArray, 0, byteArray.Length);
					memStream.Write(values, 0, values.Length);
					byteArray = memStream.ToArray();
				}
			}
			Command(cmd, byteArray);
			bool isOK = WriteCommandIsDone(cmd);

			return isOK;
		}
		
		private byte HiByte(int n) { return (byte)((n >> 8) & 0xFF); }
		private byte LoByte(int n) { return (byte)(n & 0xFF); }

		private const int
			SETDEVINFO_SET_DEVINFO = 1,
			SETDEVINFO_SET_SENSORPARAM = 2;
		private const int RSV_BYTELENGTH_DEVINFO = 4;

		//デバイス情報をセットする(memAdrs=1)
		//引数
		//memSizeKByte:	EEPROMのメモリサイズ。KByte単位
		//				例）16や32	※128Kbit=16KByte, 256Kbit=32KByte
		//pageSizeByte:	EEPROMの1ページのメモリサイズ。byte単位
		//				例）64や128	※1MbitのEEPROMで1ページ128byte。これより大きい値はないものとする
		public bool SetDeviceInfo(byte memSizeKByte, byte pageSizeByte)
		{
			byte[] byteArray = new byte[] { memSizeKByte, pageSizeByte };
			bool isOK = SendCommand(COMCmd.DEVINFO, SETDEVINFO_SET_DEVINFO, byteArray.Length, byteArray);
			if (!isOK) return false;

			byteArray = GetRsvBytes(RSV_BYTELENGTH_DEVINFO);
			return (byteArray.Length == RSV_BYTELENGTH_DEVINFO);
		}

		//測定インターバルをセットする(memAdrs=2)
		public bool SetSensorInterval(byte minute)
		{
			if (minute == 0) minute = 1;
			byte[] byteArray = new byte[] { minute };
			bool isOK = SendCommand(COMCmd.DEVINFO, SETDEVINFO_SET_SENSORPARAM, byteArray.Length, byteArray);
			if (!isOK) return false;

			byteArray = GetRsvBytes(RSV_BYTELENGTH_DEVINFO);
			return (byteArray.Length == RSV_BYTELENGTH_DEVINFO);
		}

		//現在セットされているデバイス情報を取得する(memAdrs=0)
		public bool GetDeviceInfo(out byte memSizeKByte, out byte pageSizeByte, out byte sensorInterval)
		{
			memSizeKByte = 0; pageSizeByte = 0; sensorInterval = 0;
			bool isOK = SendCommand(COMCmd.DEVINFO, 0, 0);
			if (!isOK) return false;

			byte[] byteArray = GetRsvBytes(RSV_BYTELENGTH_DEVINFO);
			if (byteArray.Length != RSV_BYTELENGTH_DEVINFO) return false;

			int idx = 0;
			int memLastAdrsHi = byteArray[idx++], memLastAdrsLo = byteArray[idx++];
			pageSizeByte = byteArray[idx++];
			sensorInterval = byteArray[idx++];
			//
			int memLastAdrs = (memLastAdrsHi << 8) + memLastAdrsLo;
			memSizeKByte = (byte)((memLastAdrs + 1) / 1024);
			
			return true;
		}

		//測定開始日時をセットする
		private const int
			SETSTARTDATE_SETPARAM = 1,
			SETSTARTDATE_GETPARAM = 0;
		private const int RSV_BYTELENGTH_STARTDATE = 6;
		public bool SetStartDate(int yyyy, int mmdd, int hhmm)
		{
			byte[] byteArray = new byte[] { HiByte(yyyy), LoByte(yyyy), HiByte(mmdd), LoByte(mmdd), HiByte(hhmm), LoByte(hhmm) };
			bool isOK = SendCommand(COMCmd.STARTDATE, SETSTARTDATE_SETPARAM, byteArray.Length, byteArray);
			if (!isOK) return false;

			byteArray = GetRsvBytes(RSV_BYTELENGTH_STARTDATE);
			return (byteArray.Length == RSV_BYTELENGTH_STARTDATE);
		}

		//現在セットされている測定開始日時を取得する
		public bool GetStartDate(out int yyyy, out int mmdd, out int hhmm)
		{
			yyyy = 0; mmdd = 0; hhmm = 0;
			bool isOK = SendCommand(COMCmd.STARTDATE, SETSTARTDATE_GETPARAM, 0);
			if (!isOK) return false;

			byte[] byteArray = GetRsvBytes(RSV_BYTELENGTH_STARTDATE);
			if (byteArray.Length != RSV_BYTELENGTH_STARTDATE) return false;

			int idx = 0, hiByte, loByte;
			hiByte = byteArray[idx++]; loByte = byteArray[idx++]; yyyy = (hiByte << 8) + loByte;
			hiByte = byteArray[idx++]; loByte = byteArray[idx++]; mmdd = (hiByte << 8) + loByte;
			hiByte = byteArray[idx++]; loByte = byteArray[idx++]; hhmm = (hiByte << 8) + loByte;

			return true;
		}

		//指定のメモリアドレスから任意のバイト数を読み込む
		public byte[] ReadData(int adrs, int length)
		{
			if (length == 0) return null;
			byte[] readDatas = new byte[length];

			Console.Write("read bytes n/{0}...", length);
			int offset = 0;
			while (offset < length)
			{
				//COMCmd.READを送信すると、まずコマンドが受け付けられた合図が返ってくる
				int reqLength = Math.Min(length - offset, DATA_BUF_SIZE);
				bool isOK = SendCommand(COMCmd.READ, adrs, reqLength);
				if (!isOK) return null;

				//続けて目的のデータが送られてくる
				byte[] byteArray = GetRsvBytes(reqLength);
				if (byteArray.Length == 0)
				{
					readDatas = byteArray;
					break;
				}
				Array.Copy(byteArray, 0, readDatas, offset, reqLength);

				offset += reqLength;
				adrs += reqLength;

				//経過表示
				Console.Write(" {0}", offset);
			}
			Console.WriteLine();

			return readDatas;
		}

		//デバイスから送出されてくるデータを受信する
		//・受信するデータの長さはあらかじめ分かっているものとする。
		private byte[] GetRsvBytes(int length)
		{
			byte[] byteArray = new byte[length];
			int offset = 0, count = byteArray.Length, readLength;
			IsComError = false;
			try
			{
				while (0 < count)
				{
					readLength = comCtrl.Read(byteArray, offset, count);
					offset += readLength;
					count -= readLength;
				}
			}
			catch
			{
				//タイムアウト発生はキャッチされる
				IsComError = true;
				return new byte[0];
			}
			finally
			{
				//PC側の受信バッファをクリアする
				//・バッファにはEND_CODEが残っている。
				RecieveBufClear();
			}

			return byteArray;
		}

		//EEPROMのメモリ領域全体を指定の値で埋める（メモリ消去）
		public bool Erace(byte data)
		{
			bool isOK = SendCommand(COMCmd.ERASE, 0, 0, new byte[]{ data });
			return isOK;
		}

	}	//class
}	//namespace
