﻿using System;
using System.IO.Ports;		//SerialPort
using System.IO;	//Stream

namespace ModifiedSerialPort
{
	//USB-シリアル変換モジュールの引っこ抜き対策をしたSerialPortクラス
	
	class MSerialPort : SerialPort
	{
		//ケーブル引っこ抜き対策のために一時保存しておくBaseStream
		private static Stream openedBaseStream;
		
		//コンストラクタ
		public MSerialPort()
			: base()
		{
			openedBaseStream = null;
		}
		
		//オープン
		public new void Open()
		{
			base.Open();
			
			//ケーブル引っこ抜き対策のため、オープンしたストリームを保存しておく
			openedBaseStream = this.BaseStream;
		}
		
		//ケーブル引っこ抜き対策込みのDispose()
		//・Dispose()はClose()からも呼ばれる。
		/*	【エラー発生の説明】
		SerialPortをオープンしたあと、USB-シリアル変換ケーブルを引っこ抜いていると、
		アプリ終了時に（Main()を抜けるときに）UnauthorizedAccessException例外が発生する。
		Open()したBaseStreamがClose()されていないので、実行環境がClose()/Dispose()しに行くが、
		当然アクセスできないので例外発生、というメカニズムである。
		Main()を抜けるときに実行環境から呼ばれるSerialPort.Dispose()で例外をキャッチする方法はあるが、
		・ダイアログを表示するなど何らかの処理が必要で、例外処理をキャンセルできないので見苦しい。
		・ダイアログを表示してもアプリはそのまま自動的に終了してしまう。
		対策：
		SerialPort.Dispose()をオーバーライドし、そこでキャッチする。
		なお、Dispose()は実行環境がオブジェクトを破棄するときだけでなく、Close()からも呼ばれる。
		*/
		/*	【発生する例外エラーの詳細】
		---------------------------------------------------------------------------------
		System.UnauthorizedAccessException はハンドルされませんでした。
		  Message="ポートへのアクセスが拒否されました。"
		  Source="System"
		  StackTrace:
			   場所 System.IO.Ports.InternalResources.WinIOError(Int32 errorCode, String str)
			   場所 System.IO.Ports.InternalResources.WinIOError()
			   場所 System.IO.Ports.SerialStream.Dispose(Boolean disposing)
			   場所 System.IO.Ports.SerialStream.Finalize()
		  InnerException: 
		---------------------------------------------------------------------------------
		SerialStreamはSerialPort.BaseStreamを指す（SerialPortはSerialStreamをラッピングしている）。
		*/
		protected override void Dispose(bool disposing)
		{
			try
			{
				if (openedBaseStream != null) openedBaseStream.Close();
				openedBaseStream = null;
			}
			catch (UnauthorizedAccessException) { }		//ケーブルを引っこ抜いていた場合、ここでキャッチされる
			
			base.Dispose(disposing);
		}
		
	}	//class
}	//namespace
