﻿/*	ATmega88系
	湿度／温度ロガー I2C-EEPROM読み書きツール
	『昼夜逆転』工作室	http://jsdiy.web.fc2.com

【更新履歴】
2013/10/08	ver 1.00	初版
*/

using System;
using System.Collections;
using System.Collections.Specialized;
using ModifiedSerialPort;

namespace I2cEepRomTool
{
	//エラーコード
	enum ErrorCode
	{
		NormalTermination = 0,
		COM_Error,
		Error = 255
	};

	//コマンドラインオプションで指定される値
	static class App
	{
		public struct TTwoByteVal { public byte Option, Current; }
		public struct TTwoWordVal { public int Option, Current; }
		public struct TStartDate { public TTwoWordVal yyyy, mmdd, hhmm; }
		
		public static string portName = null;
		public static StringCollection options = null;
		public static StringCollection commands = null;
		public static TTwoByteVal memSizeKByte, pageSizeByte, sensorInterval;
		public static bool isCurrentValueAvailable = false;
		public static TStartDate startDate;

		//初期値をセットする
		public static void InitParam()
		{
			memSizeKByte.Option = pageSizeByte.Option = sensorInterval.Option = 0;
			startDate.yyyy.Option = startDate.mmdd.Option = startDate.hhmm.Option = 0;
		}

		//オプション指定されなかったパラメータに、デバイスから取得した現在の値を代入する
		public static void CopyCurrentToOption()
		{
			if (memSizeKByte.Option == 0) memSizeKByte.Option = memSizeKByte.Current;
			if (pageSizeByte.Option == 0) pageSizeByte.Option = pageSizeByte.Current;
			if (sensorInterval.Option == 0) sensorInterval.Option = sensorInterval.Current;
		}
	}

	class Program
	{
		private const string APP_VER = "v1.00";
		private static MSerialPort comCtrl = null;
		private static string[] COMPortNameList = null;
		private static EepRomCtrl eepromCtrl = null;

		static void Main(string[] args)
		{
			//UnhandledExceptionイベント・ハンドラを登録する
			AppDomain.CurrentDomain.UnhandledException += new UnhandledExceptionEventHandler(Application_UnhandledException);

			//コマンドラインオプションを取得し、アプリのパラメータにセットする
			App.InitParam();
			CommandLineMgr cmdMgr = new CommandLineMgr();
			cmdMgr.SelectCommandLineParam(args);
			App.CopyCurrentToOption();

			//COMポート番号が指定されなかった場合は終了する
			COMPortNameList = MSerialPort.GetPortNames();
			if (string.IsNullOrEmpty(App.portName))
			{
				ShowHelp();
				AppClose(null, ErrorCode.NormalTermination);
			}

			//シリアルポートをオープンする
			//デフォルト設定：COM1, 9600bps, Parity.NONE, Data=8bit, StopBits.One
			//ReadLine()で解釈される改行コード：SerialPort.NewLine="\n"
			comCtrl = new MSerialPort();
			comCtrl.ReadTimeout = 5000;		//(ms)
			comCtrl.WriteTimeout = 5000;	//(ms)
			//comCtrl.BaudRate = 38400;	//AVR 8MHzでの代表的な安定した速度
			comCtrl.BaudRate = 4800;	//AVR 1MHzでの安定した最高速度
			comCtrl.PortName = App.portName;
			try { comCtrl.Open(); }
			catch { AppClose(comCtrl.PortName + " open error.", ErrorCode.COM_Error); }

			//デバイスの接続を確認する
			eepromCtrl = new EepRomCtrl();
			if (eepromCtrl.DeviceContact(comCtrl))
				eepromCtrl.SetSerialPort(comCtrl);
			else
				AppClose(comCtrl.PortName + " NG.", ErrorCode.COM_Error);
			GC.Collect();

			//I2C-EEPROMデバイスの現在の設定を取得する
			App.isCurrentValueAvailable =
					eepromCtrl.GetDeviceInfo(out App.memSizeKByte.Current, out App.pageSizeByte.Current, out App.sensorInterval.Current)
				&& eepromCtrl.GetStartDate(out App.startDate.yyyy.Current, out App.startDate.mmdd.Current, out App.startDate.hhmm.Current);
			if (!App.isCurrentValueAvailable)
			{
				string mes = "本システムの設定取得に失敗しました。" + Environment.NewLine +
					"デバイスが正しく接続されていることを確認してください。";
				AppClose(mes, ErrorCode.Error);
			}

			//COMポート番号以外のオプション／コマンドが指定されなかった場合、現在の設定を表示して終了する
			if (App.options.Count == 0 && App.commands.Count == 0)
			{
				ShowSystemSetting();
				AppClose(null, ErrorCode.NormalTermination);
			}

			//オプションで指定された処理を実行する
			string errMes;
			bool isOK = ActionOption(out errMes);
			if (!isOK) AppClose(errMes, ErrorCode.Error);

			//オプション指定が1つでもあった場合、更新された現在の設定を表示する
			if (0 < App.options.Count) ShowSystemSetting();

			//コマンドで指定された処理を実行する
			isOK = ActionCommand(out errMes);
			if (!isOK) AppClose(errMes, ErrorCode.Error);
			
			//アプリ正常終了（COMポートのクローズ処理込み）
			AppClose(null, ErrorCode.NormalTermination);
		}
		//end of Main()

		//未処理の例外をキャッチする
		static void Application_UnhandledException(object sender, UnhandledExceptionEventArgs e)
		{
			Console.WriteLine("予期しないエラーが発生しました。アプリを終了します。");
		}

		//アプリ終了メッセージ
		private static void AppClose(string mes, ErrorCode code)
		{
			//終了メッセージを表示する
			if (mes != null) Console.WriteLine(mes + Environment.NewLine);

			//COMポート クローズ
			try { comCtrl.Close(); }
			catch { }

			Console.WriteLine("Press any key.");
			Console.ReadKey();

			Environment.Exit((int)code);
		}

		//ヘルプ画面の見出し
		private static void ShowHelpHeader()
		{
			Console.WriteLine("湿度／温度ロガー アクセスツール {0:s}", APP_VER);
			Console.WriteLine("(C)jsdiy/2013  http://jsdiy.web.fc2.com");
			Console.WriteLine();
		}
		
		//ヘルプ画面
		private static void ShowHelp()
		{
			ShowHelpHeader();
			Console.Write(
				"[Usage]" + Environment.NewLine +
				">I2cEepRomTool.exe -com<n> [-m<n>] [-p<n>] [-t<n>] [-d<n>]" + Environment.NewLine +
				">I2cEepRomTool.exe -com<n> {r|d|e}" + Environment.NewLine +
				Environment.NewLine +
				"[Options]" + Environment.NewLine +
				"-com<n> : COMポート番号. 例)'-com4'" + Environment.NewLine +
				"-m<n>   : I2C-EEPROMのメモリ容量(Kbit単位). 例)'-m128' ←128Kbit(16KByte)" + Environment.NewLine +
				"-p<n>   : I2C-EEPROMのページサイズ(byte単位). 例)'-p64' ←1ページ64byte" + Environment.NewLine +
				"-t<n>   : 測定間隔(1分単位) 1～255分. 例)'-t30' ←30分ごとに測定する" + Environment.NewLine +
				"-d<n>   : 測定開始時刻 年月日時分. 例)'-d201310061830' ←2013/10/06-18:30" + Environment.NewLine +
				Environment.NewLine +
				"[Commands]" + Environment.NewLine +
				"r : I2C-EEPROMの内容を読み込む(データ表示)." + Environment.NewLine +
				"d : I2C-EEPROMの内容を読み込む(メモリダンプ)." + Environment.NewLine +
				"e : I2C-EEPROMの内容を消去する(0xFFで埋める). ※確認なしで即実行【注意】" + Environment.NewLine +
				Environment.NewLine +
				"オプション/コマンドなし(-com<n>のみ)で本システムの現在の設定を表示します." + Environment.NewLine +
				"");
			Console.WriteLine();
			Console.WriteLine("[Serial port]");
			for (int i = 0; i < COMPortNameList.Length; i++)
				Console.Write(COMPortNameList[i] + " ");
			Console.WriteLine();	//COMPortNameList末尾の改行
			Console.WriteLine();
		}
		
		//本システムの現在の設定値を表示する
		private static void ShowSystemSetting()
		{
			if (!App.isCurrentValueAvailable) return;

			ShowHelpHeader();
			Console.WriteLine(
				"本システムは次のように設定されています。" + Environment.NewLine +
				"  メモリ容量  : {3}Kbit ({0}KByte)" + Environment.NewLine +
				"  ページサイズ: {1}byte" + Environment.NewLine +
				"  測定間隔    : {2}分ごと" + Environment.NewLine +
				"  測定開始日時: {4}/{5:D2}/{6:D2} - {7:D2}:{8:D2}" + Environment.NewLine +
				"", App.memSizeKByte.Current, App.pageSizeByte.Current, App.sensorInterval.Current, App.memSizeKByte.Current * 8,
					App.startDate.yyyy.Current, App.startDate.mmdd.Current / 100, App.startDate.mmdd.Current % 100, App.startDate.hhmm.Current / 100, App.startDate.hhmm.Current % 100);
			Console.WriteLine();
		}

		//オプションで指定された処理を実行する
		private static bool ActionOption(out string errMes)
		{
			errMes = "";
			bool isOK = true;

			//I2C-EEPROMのデバイス情報を設定する
			//・メモリアドレスの指定方法の都合により、
			//	本機では32Kbit(=4KByte)以上、512Kbit(=64KByte)以下のI2C-EEPROMに対応する。
			//・1MbitのデバイスはIC内部的に512Kbitのデバイス2個に相当する。
			//	本機では1Mbitデバイスは非対応としているが、実際には512Kbitデバイスとして使用できる。
			if (App.options.Contains("-m") || App.options.Contains("-p"))
			{
				if (eepromCtrl.SetDeviceInfo(App.memSizeKByte.Option, App.pageSizeByte.Option))
				{
					//Currentの値を最新の内容に更新する
					App.memSizeKByte.Current = App.memSizeKByte.Option;
					App.pageSizeByte.Current = App.pageSizeByte.Option;
				}
				else
				{
					errMes += "-m -p の操作に失敗しました." + Environment.NewLine;
					isOK = false;
				}
			}

			//測定インターバルを設定する
			if (App.options.Contains("-t"))
			{
				if (eepromCtrl.SetSensorInterval(App.sensorInterval.Option))
				{
					//Currentの値を最新の内容に更新する
					App.sensorInterval.Current = App.sensorInterval.Option;
				}
				else
				{
					errMes += "-t の操作に失敗しました." + Environment.NewLine;
					isOK = false;
				}
			}

			//測定開始日時を設定する
			if (App.options.Contains("-d"))
			{
				if (eepromCtrl.SetStartDate(App.startDate.yyyy.Option, App.startDate.mmdd.Option, App.startDate.hhmm.Option))
				{
					//Currentの値を最新の内容に更新する
					App.startDate.yyyy.Current = App.startDate.yyyy.Option;
					App.startDate.mmdd.Current = App.startDate.mmdd.Option;
					App.startDate.hhmm.Current = App.startDate.hhmm.Option;
				}
				else
				{
					errMes += "-d の操作に失敗しました." + Environment.NewLine;
					isOK = false;
				}
			}
			
			return isOK;
		}

		//コマンドで指定された処理を実行する
		private static bool ActionCommand(out string errMes)
		{
			errMes = null;
			if (App.commands.Count == 0) return true;
			bool isOK;

			//I2C-EEPROMの内容を消去する(0xFFで埋める)
			if (App.commands.Contains("e"))
			{
				isOK = eepromCtrl.Erace(0xFF);
				if (!isOK) errMes = "デバイスの消去に失敗しました.";
				return isOK;
			}

			//I2C-EEPROMの内容を読み込む
			int length = App.memSizeKByte.Current * 1024;
			//int length = 1 * 1024;	//debug
			byte[] datas = eepromCtrl.ReadData(0, length);
			isOK = (datas != null && datas.Length == length);
			if (!isOK)
			{
				errMes = "データの読み込みに失敗しました.";
				return false;
			}

			DataDisplay dataDisplay = new DataDisplay();

			//I2C-EEPROMの内容を読み込む(データ表示)
			if (App.commands.Contains("r")) dataDisplay.RecordForm(datas);

			//I2C-EEPROMの内容を読み込む(メモリダンプ)
			if (App.commands.Contains("d")) dataDisplay.DumpListForm(datas);
			
			return true;
		}

	}	//class

	class CommandLineMgr
	{
		private void SafeScAdd(StringCollection sc, string str)
		{
			if (!sc.Contains(str)) sc.Add(str);
		}

		//コマンドライン引数をオプション／コマンドに分ける
		//・オプションの場合、値を解釈してアプリのパラメータにセットする。
		public void SelectCommandLineParam(string[] cmdLineArgs)
		{
			App.options = new StringCollection();
			App.commands = new StringCollection();
			if (cmdLineArgs.Length == 0) return;

			foreach (string cmdLineArg in cmdLineArgs)
			{
				if (string.IsNullOrEmpty(cmdLineArg)) continue;
				string arg = cmdLineArg.Trim().ToLower();
				if (string.IsNullOrEmpty(arg)) continue;

				if (arg.StartsWith("-"))
				{
					string optionCode = SetAppOptionParam(arg);
					if (optionCode != null) SafeScAdd(App.options, optionCode);
				}
				else if (arg == "r" || arg == "d" || arg == "e")
				{
					SafeScAdd(App.commands, arg);
				}
			}
		}

		//コマンドラインオプションを解釈し、アプリのパラメータをセットする
		//・オプション「-xNNN」の「NNN」を数値としてアプリのパラメータにセットし、「-x」を返す。
		//	返した値は配列に保存される。その配列は「コマンドラインに'-x'オプションは含まれていたか？」の判定に使われる。
		//引数
		//	option:	数字付きのコマンドラインオプションの文字列「-xNNN」
		//戻り値
		//	数字を除いたコマンドラインオプションの文字列「-x」
		private string SetAppOptionParam(string option)
		{
			byte bNum;
			int iNum;
			
			if (option.StartsWith("-com"))
			{
				App.portName = "COM" + option.Substring(4);
				return null;
			}
			
			if (option.StartsWith("-m"))	//Kbit単位
			{
				if (ParseInt(option.Substring(2), out iNum))
				{
					bNum = (byte)(iNum / 8);	//KByte単位に換算
					if (!(0 < bNum && bNum <= 64)) return null;	//64KByteまで
					App.memSizeKByte.Option = bNum;
					return option.Substring(0, 2);
				}
			}
			
			if (option.StartsWith("-p"))
			{
				if (ParseByte(option.Substring(2), out bNum) && (0 < bNum))
				{
					App.pageSizeByte.Option = bNum;
					return option.Substring(0, 2);
				}
			}
			
			if (option.StartsWith("-t"))
			{
				if (ParseByte(option.Substring(2), out bNum) && (0 < bNum))
				{
					App.sensorInterval.Option = bNum;
					return option.Substring(0, 2);
				}
			}

			if (option.StartsWith("-d"))	//yyyymmddhhmm:年月日時分
			{
				int yyyy, mmdd, hhmm;
				if (ParseInt(option.Substring(2+0, 4), out yyyy) &&
					ParseInt(option.Substring(2+4, 4), out mmdd) &&
					ParseInt(option.Substring(2+8, 4), out hhmm))
				{
					App.startDate.yyyy.Option = yyyy;
					App.startDate.mmdd.Option = mmdd;
					App.startDate.hhmm.Option = hhmm;
					return option.Substring(0, 2);
				}
			}
			
			return null;
		}

		private bool ParseByte(string strNum, out byte byteNum)
		{
			byteNum = 0x00;
			bool isOK = true;
			try { byteNum = byte.Parse(strNum); }
			catch { isOK = false; }
			return isOK;
		}

		private bool ParseInt(string strNum, out int intNum)
		{
			intNum = 0;
			bool isOK = true;
			try { intNum = int.Parse(strNum); }
			catch { isOK = false; }
			return isOK;
		}

	}	//class

	class DataDisplay
	{
		//レコード形式で表示する（やっつけ仕事もいいところ）
		public void RecordForm(byte[] datas)
		{
			//{ [0]indexLo, [1]indexHi, [2]経過_日数, [3]経過_時間, [4]経過_分, [5]湿度, [6]温度 }の繰り返し
			SortedList sl = new SortedList();
			int index;
			int[] values;
			for (int i = 0; i < datas.Length && i + 7 <= datas.Length; i += 7)
			{
				index = (datas[i + 1] << 8) + datas[i + 0];	//リトルエンディアンで16bit値を復元する
				values = new int[] { datas[i + 2], datas[i + 3], datas[i + 4],	//日,時,分
					datas[i + 5], datas[i + 6] };	//湿度, 温度

				if (sl.ContainsKey(index) || values[0] == 0xFF) continue;
				sl.Add(index, values);
			}

			Console.WriteLine("測定開始日時: {0}/{1:D2}/{2:D2} - {3:D2}:{4:D2}",
				App.startDate.yyyy.Current, App.startDate.mmdd.Current / 100, App.startDate.mmdd.Current % 100, App.startDate.hhmm.Current / 100, App.startDate.hhmm.Current % 100);
			Console.WriteLine("idx: 経過日数-時間-分 相対湿度%RH 温度'C");
			for (int i = 0; i < sl.Count; i++)
			{
				index = (int)sl.GetKey(i);
				values = (int[])sl[index];
				Console.WriteLine("{0}: {1}-{2:D2}-{3:D2} {4}%RH {5}'C", index, values[0], values[1], values[2], values[3], values[4]);
			}
		}

		//ダンプリスト形式で表示する
		public void DumpListForm(byte[] datas)
		{
			for (int i = 0; i < datas.Length; i++)
			{
				if (i % 16 != 0) Console.Write(" ");
				Console.Write("{0:X2}", datas[i]);
				if (i % 16 == 15) Console.WriteLine();
			}
		}

	}	//class

}	//namespace
